/***************************** Include Files *********************************/
#include <stdio.h>
#include <string.h>

#include "lwip/sockets.h"
#include "lwip/inet.h"
#include "netif/xadapter.h"
#include "lwipopts.h"
#include "xil_printf.h"
#include "FreeRTOS.h"
#include "task.h"
#include "stdbool.h"
#include "qspi.h"
#include "xil_types.h"
#include "image_mover.h"
#include "xstatus.h"

#include "http_server.h"
/************************** Constant Definitions *****************************/
#define HTTP_PORT		    (80)
#define BOOT_IMAGE_MAX_SIZE (0x400000) // 4 MB Max Boot Image Size
#define RECV_BUF_SIZE       (0x800)    // 2 MB Max Packet Size
/**************************** Type Definitions *******************************/

/************************** Function Prototypes ******************************/
void process_http_request(int sd);

int generate_http_response(int sd, HttpRequest_t *Request);

/****************** proces_bootimage_update **********************************/
void process_bootimage_update(const u32 BootImageSize);
void process_goldenimage_update(const u32 BootImageSize);
/*****************************************************************************/

u8 *GetHttpDataStart(u8 *RecvData);
void ProcessRequestHeader(const u8 *HttpHeader, HttpRequest_t *Request);
void GetRequestMethodAndCommand(const u8 *HttpHeader, HttpRequest_t *Request);
u32 GetContentLength(const u8 *HttpHeader);
WebCommand_t GetWebCommand(const u8 *HttpHeader);

/************************** Variable Definitions *****************************/
u8  DataBuf[BOOT_IMAGE_MAX_SIZE];
u8 *DataBufPtr = &DataBuf[0];

void http_server_thread()
{
    int sd, new_sd;
    int size;
    struct sockaddr_in address, remote;

    memset ((void *)(&address), 0x0, sizeof(address));

    /* create a TCP socket */
    if ((sd = lwip_socket(AF_INET, SOCK_STREAM, 0)) < 0) {
        xil_printf("\r\n%s: Failed to create socket.\r\n", __FUNCTION__);
        return;
    }

    /* bind to port 80 at any interface */
    address.sin_family = AF_INET;
    address.sin_port = htons(HTTP_PORT);
    address.sin_addr.s_addr = INADDR_ANY;

    if (lwip_bind(sd, (struct sockaddr *)&address, sizeof(address)) < 0) {
        xil_printf("\r\n%s: Failed to bind at port '%.d'\r\n", __FUNCTION__,
                HTTP_PORT);
        return;
    }

    /* listen for incoming connections */
    if (lwip_listen(sd, 0)) {
        xil_printf("\r\n%s: Failed to listen.\r\n", __FUNCTION__);
        return;
    }

    while (1) {
        if ((new_sd = lwip_accept(sd, (struct sockaddr *)&remote,
            (socklen_t *)&size)) > 0) {
            sys_thread_new("http_thread", process_http_request, (void *)new_sd,
                    5120, DEFAULT_THREAD_PRIO);
        }
    }
}

void process_http_request(int sd)
{
    int n = 0;
    int ByteCount = 0;
    u32 Status;
    u8 recv_buf[RECV_BUF_SIZE];

    /* Http request related variables */
    HttpRequest_t   Request;
    bool            RequestHeaderReceived = false;
    u32             RequestHeaderSize;

    u8             *DataStartPtr;
    u32             DataCount = 0;

    while (1) {

        /* Read data from socket until whole request header is received */
        if (!RequestHeaderReceived) {

            if ((DataStartPtr = GetHttpDataStart(recv_buf)) == NULL) {

                if ((n = read(sd, recv_buf + n, RECV_BUF_SIZE - ByteCount))
                        <= 0) {
                    xil_printf("\r\n%s: Error reading from socket '%d'. "
                            "Closing connection.\r\n", __FUNCTION__, sd);
                    break;
                }
                ByteCount += n;
            } else {

                RequestHeaderReceived = true;
                RequestHeaderSize = (u32)(DataStartPtr) - (u32)(&recv_buf[0]);

                /* Http request header dump */
                xil_printf("---------- Request Header ----------\r\n\r\n");
                for (u32 Index = 0; Index < RequestHeaderSize; Index++) {
                    xil_printf("%c", recv_buf[Index]);
                }

                /*
                 * Process http request header to determine method, command and
                 * content length, if the request have some
                 */
                ProcessRequestHeader(recv_buf, &Request);

                /* Move http data from current tcp packet to data buffer */
                memcpy((void *)DataBufPtr, (const void *)(recv_buf),
                        (size_t)(n - RequestHeaderSize));

                DataBufPtr += (n - RequestHeaderSize);
                DataCount += (n - RequestHeaderSize);
            }
        } else if (DataCount < Request.ContentLength) {

            /*
             * Continue reading from socket until whole http content is received
             */
            if ((n = read(sd, recv_buf, RECV_BUF_SIZE)) <= 0) {
                xil_printf("\r\n%s: Error reading from socket '%d'."
                        "Closing connection.\r\n", __FUNCTION__, sd);
                break;
            }

            memcpy((void *)DataBufPtr, (const void *)recv_buf, (size_t)n);

            DataBufPtr += n;
            DataCount += n;

        } else {

            xil_printf("Received %d bytes of data. Sending Response.\r\n", DataCount);

            Status = generate_http_response(sd, &Request);
            if (Status != XST_SUCCESS) {
                xil_printf("\r\n%s: Error responding to request. "
                        "Closing connection.\r\n", __FUNCTION__);
                break;
            }

            if (Request.Method == HTTP_POST) {
                u32 ts;

                ts = GetBootImageTimestamp(DataBuf);

                xil_printf("\r\nTimestamp of received image: 0x%08X\r\n", ts);

                switch (Request.Command) {
                case UPDATE_MULTIBOOT_CMD:
                    process_bootimage_update(DataCount);
                    break;
                case UPDATE_GOLDEN_CMD:
                    process_goldenimage_update(DataCount);
                    break;
                default:
                    break;
                }
            }
            break;
        }
    }
    /* close connection */
    xil_printf("Closing connection...\r\n");
    close(sd);
    vTaskDelete(NULL);
}
void process_bootimage_update(const u32 BootImageSize)
{
    time_t TimestampBootImageA, TimestampBootImageB;
    u32 UpdateAddress;

    if (FlashIsBootImageValid(BOOT_IMAGE_A_ADDRESS)) {

        if (FlashIsBootImageValid(BOOT_IMAGE_B_ADDRESS)) {

            TimestampBootImageA = FlashGetBootImageTimestamp(BOOT_IMAGE_A_ADDRESS);
            TimestampBootImageB = FlashGetBootImageTimestamp(BOOT_IMAGE_B_ADDRESS);

            xil_printf("Timestamp of boot image on position A: %08X\r\n", TimestampBootImageA);
            xil_printf("Timestamp of boot image on position B: %08X\r\n", TimestampBootImageB);

            if (TimestampBootImageA > TimestampBootImageB) {
                UpdateAddress = BOOT_IMAGE_B_ADDRESS;
                xil_printf("Detected older boot image B. Update address will be B: 0x%08X\r\n", BOOT_IMAGE_B_ADDRESS);
            } else {
                xil_printf("Detected older boot image A. Update address will be A: 0x%08X\r\n", BOOT_IMAGE_A_ADDRESS);
                UpdateAddress = BOOT_IMAGE_A_ADDRESS;
            }
        } else {
            xil_printf("Detected invalid boot image B. Update address will be B: 0x%08X\r\n", BOOT_IMAGE_B_ADDRESS);
            UpdateAddress = BOOT_IMAGE_B_ADDRESS;
        }
    } else {
        xil_printf("Detected invalid boot image A. Update address will be A: 0x%08X\r\n", BOOT_IMAGE_A_ADDRESS);
        UpdateAddress = BOOT_IMAGE_A_ADDRESS;
    }
    // UpdateAddress = 0x0;
    FlashErase(UpdateAddress, BOOT_IMAGE_MAX_SIZE); /* Erase whole boot image sector */
    xil_printf("Erasing Done.\r\n");
    FlashWrite(UpdateAddress, DataBuf, BootImageSize);
    xil_printf("Writing Done.\r\n");
    FlashValidateBootImage(UpdateAddress);
    xil_printf("Validating Done.\r\n");
}

void process_goldenimage_update(const u32 BootImageSize)
{
    u32 UpdateAddress = GOLDEN_IMAGE_ADDRESS;

    /* Erase whole boot image sector */
    FlashErase(UpdateAddress, BOOT_IMAGE_MAX_SIZE);
    xil_printf("Erasing Done.\r\n");
    FlashWrite(UpdateAddress, DataBuf, BootImageSize);
    xil_printf("Writing Done.\r\n");
    FlashValidateBootImage(UpdateAddress);
    xil_printf("Validating Done.\r\n");
}
