/***************************** Include Files *********************************/
#include "image_mover.h"
#include "xil_printf.h"
#include "xreg_cortexa9.h"
#include "md5.h"
#include "string.h"
#include "time.h"
#include "xstatus.h"
#include "qspi.h"
/************************** Constant Definitions *****************************/

/* We are 32-bit machine */
#define MAXIMUM_IMAGE_WORD_LEN 0x40000000
#define MD5_CHECKSUM_SIZE   16

/**************************** Type Definitions *******************************/

/***************** Macros (Inline Functions) Definitions *********************/

/************************** Function Prototypes ******************************/

/************************** Variable Definitions *****************************/
u8 BootImageBuffer[200000];
u8 PartitionBuffer[200000];
/*
 * Partition information flags
 */

u8 PartitionChecksumFlag;

u32 GetPartitionCount(u8 *BootImageAddr)
{
    u32 PartitionCount = 0;
    u32 ImagePartitionCount;
    u32 ImageNum;
    u32 ImageCount;
    u32 ImageHeaderTableOffset;     // Byte Offset
    u32 ImageHeaderOffset;

    /*
     * Get Image Header Table Offset (Byte Offset)
     */
    ImageHeaderTableOffset = *(u32*)(BootImageAddr + IMAGE_HDR_OFFSET);

    /*
     * Get Image Count
     */
    ImageCount = *(u32*)(BootImageAddr + ImageHeaderTableOffset + IMAGE_HDR_COUNT_OFFSET);
    xil_printf("Total Number of Images in Boot Image: %d\r\n", ImageCount);
    /*
     * Get First Image Header Offset (Word Offset)
     */
    ImageHeaderOffset = (*(u32*)(BootImageAddr + ImageHeaderTableOffset
                                          + FIRST_IMAGE_HDR_OFFSET)) << 2;
    xil_printf("First Image Header Offset: 0x%08x\r\n", ImageHeaderOffset);

    for (ImageNum = 0; ImageNum < ImageCount; ImageNum++) {

        /*
         * Get Partition Count From Image Header
         */
        ImagePartitionCount = *(u32*)(BootImageAddr + ImageHeaderOffset
                                            + PARTITION_COUNT_OFFSET);
        xil_printf("Number of Partitions Associated With %d. Image: %d\r\n",
        		ImageNum + 1, ImagePartitionCount);

        PartitionCount += ImagePartitionCount;
        /*
         * Get Next Image Header Offset (Word Offset)
         */
        ImageHeaderOffset = (*(u32*)(BootImageAddr + ImageHeaderOffset)) << 2;
    }

    xil_printf("Total Number of Partitions in Boot Image: %d\r\n", PartitionCount);
    
    return PartitionCount;
}

u32 ValidatePartition(u8* BootImageAddr, PartHeader *PartitionHeader)
{
    u32 PartitionChecksumOffset;
    u32 PartitionStartOffset;
    u32 PartitionLength;
    u32 Index;
    u8  Checksum[MD5_CHECKSUM_SIZE];
    u8  CalcChecksum[MD5_CHECKSUM_SIZE];
    
    PartitionChecksumOffset = (PartitionHeader->CheckSumOffset) << 2; // Word offset
    PartitionStartOffset    = (PartitionHeader->PartitionStart) << 2;
    PartitionLength         = (PartitionHeader->PartitionWordLen) << 2;
    
    memcpy((void *)(Checksum),
           (const void *)(BootImageAddr + PartitionChecksumOffset),
           (size_t)(MD5_CHECKSUM_SIZE));
    
    md5(BootImageAddr + PartitionStartOffset, PartitionLength, CalcChecksum, 0);
    
    for (Index = 0; Index < MD5_CHECKSUM_SIZE; Index++) {
        if(Checksum[Index] != CalcChecksum[Index]) {
            return XST_FAILURE;
        }
    }
    
    return XST_SUCCESS; 
}

u32 FlashValidatePartition(u32 FlashAddr, PartHeader *PartitionHeader)
{
    u32 PartitionChecksumOffset;
    u32 PartitionStartOffset;
    u32 PartitionLength;
    u32 Index;
    u8  Checksum[MD5_CHECKSUM_SIZE];
    u8  CalcChecksum[MD5_CHECKSUM_SIZE];

    PartitionChecksumOffset = (PartitionHeader->CheckSumOffset) << 2; // Word offset
    PartitionStartOffset    = (PartitionHeader->PartitionStart) << 2;
    PartitionLength         = (PartitionHeader->PartitionWordLen) << 2;

    FlashRead(Checksum, FlashAddr + PartitionChecksumOffset, MD5_CHECKSUM_SIZE);
    FlashRead(PartitionBuffer, FlashAddr + PartitionStartOffset, PartitionLength);

    xil_printf("Flash Partition Checksum: ");
    for (Index = 0; Index < MD5_CHECKSUM_SIZE; Index++) {
        xil_printf("%02x", Checksum[Index]);
    }
    xil_printf("\r\n");

    md5(PartitionBuffer, PartitionLength, CalcChecksum, 0);

    xil_printf("Calculated Partition Checksum: ");
    for (Index = 0; Index < MD5_CHECKSUM_SIZE; Index++) {
        xil_printf("%02x", CalcChecksum[Index]);
    }
    xil_printf("\r\n");

    for (Index = 0; Index < MD5_CHECKSUM_SIZE; Index++) {
        if(Checksum[Index] != CalcChecksum[Index]) {
            return XST_FAILURE;
        }
    }

    return XST_SUCCESS;
}

u32 FlashValidateFsblPartition(u32 FlashAddr, PartHeader *PartitionHeader)
{
    u32 PartitionChecksumOffset;
    u32 PartitionStartOffset;
    u32 PartitionLength;
    u32 Index;
    u8  Checksum[MD5_CHECKSUM_SIZE];
    u8  CalcChecksum[MD5_CHECKSUM_SIZE];

    PartitionStartOffset    = (PartitionHeader->PartitionStart) << 2;
    PartitionLength         = (PartitionHeader->PartitionWordLen) << 2;

    FlashRead(Checksum, FlashAddr + FSBL_CHECKSUM_OFFSET, MD5_CHECKSUM_SIZE);
    FlashRead(PartitionBuffer, FlashAddr + PartitionStartOffset, PartitionLength);

    xil_printf("Flash FSBL Checksum: ");
    for (Index = 0; Index < MD5_CHECKSUM_SIZE; Index++) {
        xil_printf("%02x", Checksum[Index]);
    }
    xil_printf("\r\n");

    md5(PartitionBuffer, PartitionLength, CalcChecksum, 0);

    xil_printf("Calculated FSBL Checksum: ");
    for (Index = 0; Index < MD5_CHECKSUM_SIZE; Index++) {
        xil_printf("%02x", CalcChecksum[Index]);
    }
    xil_printf("\r\n");

    for (Index = 0; Index < MD5_CHECKSUM_SIZE; Index++) {
        if(Checksum[Index] != CalcChecksum[Index]) {
            return XST_FAILURE;
        }
    }

    return XST_SUCCESS;
}


u32 ValidateFsblPartition(u8* BootImageAddr, PartHeader *PartitionHeader)
{
    u32 PartitionStartOffset;
    u32 PartitionLength;
    u32 Index;
    u8  Checksum[MD5_CHECKSUM_SIZE];
    u8  CalcChecksum[MD5_CHECKSUM_SIZE];
    
    PartitionStartOffset    = (PartitionHeader->PartitionStart) << 2;
    PartitionLength         = (PartitionHeader->PartitionWordLen) << 2;
    
    memcpy((void *)(Checksum),
           (const void *)(BootImageAddr + FSBL_CHECKSUM_OFFSET),
           (size_t)(MD5_CHECKSUM_SIZE));
    
    xil_printf("FSBL Checksum: ");
    for (Index = 0; Index < MD5_CHECKSUM_SIZE; Index++) {
    	xil_printf("%02x", Checksum[Index]);
    }
    xil_printf("\r\n");

    md5(BootImageAddr + PartitionStartOffset, PartitionLength, CalcChecksum, 0);
    
    xil_printf("FSBL Calculated Checksum: ");
    for (Index = 0; Index < MD5_CHECKSUM_SIZE; Index++) {
    	xil_printf("%02x", CalcChecksum[Index]);
    }
    xil_printf("\r\n");

    for (Index = 0; Index < MD5_CHECKSUM_SIZE; Index++) {
        if(Checksum[Index] != CalcChecksum[Index]) {
            return XST_FAILURE;
        }
    }
    
    return XST_SUCCESS; 
}

void GetPartitionHeaderInfo(u8 *BootImageAddr, u32 PartitionCount, PartHeader *PartitionHeader)
{
    u32 PartitionHeaderTableOffset;
    
    PartitionHeaderTableOffset = *(u32*)(BootImageAddr + IMAGE_PHDR_OFFSET);
    
    memcpy((void *)(PartitionHeader),                               /* Destination */
           (const void *)(BootImageAddr + PartitionHeaderTableOffset),  /* Source */
           (size_t)(sizeof(PartHeader)*PartitionCount));            /* Number of Bytes */
}

u32 ValidateBootImage(u8 *BootImageAddr)
{
    u32 Status;
    u32 PartitionCount;
    u32 PartitionNum;
    PartHeader PartitionHeader[MAX_PARTITION_NUMBER];
    
    PartitionCount = GetPartitionCount(BootImageAddr);
    GetPartitionHeaderInfo(BootImageAddr, PartitionCount, PartitionHeader);
    
    for (PartitionNum = 0; PartitionNum < PartitionCount; PartitionNum++) {
        
    	xil_printf("Validating partition n. %d header.\r\n", PartitionNum + 1);
        Status = ValidatePartitionHeader(&PartitionHeader[PartitionNum]);
        if (Status != XST_SUCCESS) {
            xil_printf("Validation of partition header number %d failed\r\n", PartitionNum + 1);
            return XST_FAILURE;
        }
        xil_printf("Validation of partition n. %d header successful..\r\n", PartitionNum + 1);
        if (PartitionNum == 0) {
        	xil_printf("Validating FSBl partition.\r\n");
            Status = ValidateFsblPartition(BootImageAddr, &PartitionHeader[PartitionNum]);
        } else {
        	xil_printf("Validating partition n. %d.\r\n", PartitionNum + 1);
            Status = ValidatePartition(BootImageAddr, &PartitionHeader[PartitionNum]);
        }
        if (Status != XST_SUCCESS) {
            xil_printf("Validation of partition number %d failed.\r\n", PartitionNum + 1);
            return XST_FAILURE;
        }
    }
    return XST_SUCCESS;
}

u32 ValidatePartitionHeader(PartHeader *PartitionHeader)
{
    struct HeaderArray *HeaderArrayPtr;
    u32 Checksum = 0;
    u32 HeaderWord;
    
    HeaderArrayPtr = (struct HeaderArray *)PartitionHeader;
    
    for (HeaderWord = 0; HeaderWord < PARTITION_HDR_CHECKSUM_WORD_COUNT; HeaderWord++) {
        /*
         * Read and sum first 15 words from the partition header
         */
        Checksum += HeaderArrayPtr->Fields[HeaderWord];
	}
    
    /*
	 * Invert checksum, last bit of error checking
	 */
	Checksum ^= 0xFFFFFFFF;
    
    /*
	 * Validate the checksum
	 */
	if (HeaderArrayPtr->Fields[PARTITION_HDR_CHECKSUM_WORD_COUNT] != Checksum) {
	    xil_printf("Error: Partition Header Checksum 0x%8.8lx != 0x%8.8lx\r\n",
            Checksum, HeaderArrayPtr->Fields[PARTITION_HDR_CHECKSUM_WORD_COUNT]);
		return XST_FAILURE;
	}

	return XST_SUCCESS;
}

u32 GetBootImageTimestamp(const u8 *BootImageAddr)
{
    u32 Timestamp;
    time_t tm;

    Timestamp = *(u32*)(BootImageAddr + IMAGE_TIMESTAMP_OFFSET);
    tm = Timestamp;
    xil_printf("Boot Image Timestamp: %s\r\n", ctime (&tm));
    return Timestamp;
}

u32 ValidateBootRomHeader(const BootRomHeader_t *BootRomHeader)
{
    struct BootRomHeaderArray *BootRomHeaderArrayPtr;
    u32 Checksum;
    u32 HeaderWord;

    BootRomHeaderArrayPtr = (struct BootRomHeaderArray *)BootRomHeader;

    for (HeaderWord = 0; HeaderWord < BOOTROM_HDR_CHECKSUM_WORD_COUNT; HeaderWord++) {
        /*
         * Read and sum first 15 words from the partition header
         */
        Checksum += BootRomHeaderArrayPtr->Fields[HeaderWord];
    }

    /*
     * Invert checksum, last bit of error checking
     */
    Checksum ^= 0xFFFFFFFF;

    /*
     * Validate the checksum
     */
    if (BootRomHeaderArrayPtr->Fields[PARTITION_HDR_CHECKSUM_WORD_COUNT] != Checksum) {
        xil_printf("Error: Partition Header Checksum 0x%8.8lx != 0x%8.8lx\r\n",
            Checksum, BootRomHeaderArrayPtr->Fields[PARTITION_HDR_CHECKSUM_WORD_COUNT]);
        return XST_FAILURE;
    }

    return XST_SUCCESS;

}

time_t FlashGetBootImageTimestamp(const u32 FlashAddr)
{
    time_t Timestamp;

    FlashRead((u8 *)(&Timestamp), FlashAddr + IMAGE_TIMESTAMP_OFFSET, 4);

    xil_printf("Hex Timestamp: 0x%08X\r\n", Timestamp);
    xil_printf("Boot Image Timestamp: %s\r\n", ctime (&Timestamp));

    return Timestamp;
}

u32 FlashIsBootImageValid(const u32 FlashAddr)
{
    u32 Id;
    u32 ImageValidityWord;
    u8 Header[0x68];

    FlashRead(&Header[0], FlashAddr, 0x68);

    xil_printf("Flash Header Dump:\r\n\r\n");
    for (u32 row = 0x0; row < 0x68; row = row + 0x8) {
        for (u32 col = 0x0; col < 8; col++) {
            xil_printf("%02X", Header[row + col]);
        }
        xil_printf("\r\n");
    }
    xil_printf("\r\n");

    FlashRead((u8 *)(&Id), FlashAddr + IMAGE_IDENT_OFFSET, 4);
    FlashRead((u8 *)(&ImageValidityWord), FlashAddr + IMAGE_VALIDITY_WORD_OFFSET, 4);

    xil_printf("Word: 0x%08X Id: 0x%08X\r\n", ImageValidityWord, Id);

    if ((ImageValidityWord == IMAGE_VALID) && (Id == IMAGE_IDENT)) {
        xil_printf("Valid boot image found on address 0x%08X.\r\n"
                "Image Validity Word: 0x%08X Id: 0x%08X\r\n", FlashAddr, ImageValidityWord, Id);
        return 1;
    } else {
        xil_printf("Valid boot image not found on address 0x%08X.\r\n"
                        "Image Validity Word: 0x%08X Id: 0x%08X\r\n", FlashAddr, ImageValidityWord, Id);
        return 0;
    }
}

int FlashValidateBootImage (const u32 FlashAddr)
{
    u32 Status;
    u32 PartitionCount;
    u32 PartitionNum;

    PartHeader PartitionHeader[MAX_PARTITION_NUMBER];
    BootRomHeader_t BootRomHeader;
    u32 ImageValidityWord[1] = {IMAGE_VALID};
    u8 Header[0x68];

    /*
     * Get the total number of partitions in a boot image by obtaining
     * the totoal number of images from a image header table. Program then
     * goes through all the images, counting the corresponding partitions.
     */
    PartitionCount = FlashGetPartitionCount(FlashAddr);

    /* Read a boot image header */
    FlashGetBootRomHeader(FlashAddr, &BootRomHeader);

    /* Read a partition header of every partition in the boot image */
    FlashGetPartitionHeaderInfo(FlashAddr, PartitionCount, PartitionHeader);

    /*
     * Go through all the partition, validate the partition header If the
     * partition header is valid, validate partition data. A special function
     * is used to validate a FSBL partition because a FSBL partition checksum
     * is stored in UDF of the boot image header, unlike a common partitions
     * checksums, which are stored at the end of the boot image file.
     */
    for (PartitionNum = 0; PartitionNum < PartitionCount; PartitionNum++) {

        xil_printf("Validating partition n. %d header.\r\n", PartitionNum + 1);
        Status = ValidatePartitionHeader(&PartitionHeader[PartitionNum]);
        if (Status != XST_SUCCESS) {
            xil_printf("Validation of partition header"
                    " n. %d failed..\r\n", PartitionNum + 1);
            return XST_FAILURE;
        }
        xil_printf("Validation of partition"
                " n. %d header successful..\r\n", PartitionNum + 1);
        if (PartitionNum == 0) {
            xil_printf("Validating FSBl partition.\r\n");
            Status = FlashValidateFsblPartition(FlashAddr,
                    &PartitionHeader[PartitionNum]);
        } else {
            xil_printf("Validating partition n. %d.\r\n", PartitionNum + 1);
            Status = FlashValidatePartition(FlashAddr,
                    &PartitionHeader[PartitionNum]);
        }
        if (Status != XST_SUCCESS) {
            xil_printf("Validation of partition number"
                    " %d failed.\r\n", PartitionNum + 1);
            return XST_FAILURE;
        }
    }

    FlashRead(&Header[0], FlashAddr, 0x68);

    xil_printf("Flash Header Dump:\r\n\r\n");
    for (u32 row = 0x0; row < 0x68; row = row + 0x8) {
        for (u32 col = 0x0; col < 8; col++) {
            xil_printf("%02X", Header[row + col]);
        }
        xil_printf("\r\n");
    }
    xil_printf("\r\n");

    /*
     * Write 0xfffffffe at 0x60 offset address in flash to mark the
     * boot image valid.
     */
    FlashWriteWord(FlashAddr + IMAGE_VALIDITY_WORD_OFFSET, IMAGE_VALID);

    FlashRead(&Header[0], FlashAddr, 0x68);

    xil_printf("Flash Header Dump:\r\n\r\n");
    for (u32 row = 0x0; row < 0x68; row = row + 0x8) {
        for (u32 col = 0x0; col < 8; col++) {
            xil_printf("%02X", Header[row + col]);
        }
        xil_printf("\r\n");
    }
    xil_printf("\r\n");

    return XST_SUCCESS;
}

void FlashGetBootRomHeader(const u32 FlashAddr, BootRomHeader_t *BootRomHeader)
{
    FlashRead((u8 *)BootRomHeader, FlashAddr + BOOTROM_HDR_START_OFFSET, 44);
}

u32 FlashGetPartitionCount(const u32 FlashAddr)
{
    u32 PartitionCount = 0;
    u32 ImagePartitionCount;
    u32 ImageNum;
    u32 ImageCount;
    u32 ImageHeaderTableOffset;     // Byte Offset
    u32 ImageHeaderOffset;

    /*
     * Get Image Header Table Offset (Byte Offset)
     */
    FlashRead((u8 *)(&ImageHeaderTableOffset), FlashAddr  +IMAGE_HDR_OFFSET, 4);

    /*
     * Get Image Count
     */
    FlashRead((u8 *)(&ImageCount), FlashAddr + ImageHeaderTableOffset + IMAGE_HDR_COUNT_OFFSET, 4);

    xil_printf("Total Number of Images in Boot Image: %08x\r\n", ImageCount);
    /*
     * Get First Image Header Offset (Word Offset)
     */
    FlashRead((u8 *)(&ImageHeaderOffset),
            FlashAddr  + ImageHeaderTableOffset + FIRST_IMAGE_HDR_OFFSET, 4);
    ImageHeaderOffset <<= 2;

    xil_printf("First Image Header Offset: 0x%08x\r\n", ImageHeaderOffset);

    for (ImageNum = 0; ImageNum < ImageCount; ImageNum++) {

        /*
         * Get Partition Count From Image Header
         */
        FlashRead((u8 *)(&ImagePartitionCount),
                FlashAddr + ImageHeaderOffset + PARTITION_COUNT_OFFSET, 4);

        xil_printf("Number of Partitions Associated With %d. Image: %d\r\n",
                ImageNum + 1, ImagePartitionCount);

        PartitionCount += ImagePartitionCount;
        /*
         * Get Next Image Header Offset (Word Offset)
         */
        FlashRead((u8 *)(&ImageHeaderOffset), FlashAddr + ImageHeaderOffset, 4);
        ImageHeaderOffset <<= 2;
    }

    xil_printf("Total Number of Partitions in Boot Image: %d\r\n", PartitionCount);

    return PartitionCount;
}

void FlashGetPartitionHeaderInfo(const u32 FlashAddr, u32 PartitionCount, PartHeader *PartitionHeader)
{
    u32 PartitionHeaderTableOffset;

    FlashRead((u8 *)(&PartitionHeaderTableOffset), FlashAddr + IMAGE_PHDR_OFFSET, 4);
    FlashRead((u8 *)(PartitionHeader), FlashAddr + PartitionHeaderTableOffset,
            (sizeof(PartHeader)*PartitionCount));
}


