#ifndef ___IMAGE_MOVER_H___
#define ___IMAGE_MOVER_H___

/***************************** Include Files *********************************/
#include "xil_types.h"
#include "time.h"
/************************** Constant Definitions *****************************/

/* Flash Memory Map */
#define GOLDEN_IMAGE_ADDRESS             (0x0)
#define BOOT_IMAGE_A_ADDRESS             (0x400000)
#define BOOT_IMAGE_B_ADDRESS             (0x800000)


#define PARTITION_NUMBER_SHIFT	24
#define MAX_PARTITION_NUMBER	(0xE)

/* Boot Image Header defines */
#define IMAGE_HDR_OFFSET			0x098	/* Pointer to Image Header Table */
#define IMAGE_PHDR_OFFSET			0x09C	/* Ptr to Partition Header Table */
#define IMAGE_HEADER_SIZE			(64)
#define IMAGE_HEADER_TABLE_SIZE		(64)
#define TOTAL_PARTITION_HEADER_SIZE	(MAX_PARTITION_NUMBER * IMAGE_HEADER_SIZE)
#define TOTAL_IMAGE_HEADER_SIZE		(MAX_PARTITION_NUMBER * IMAGE_HEADER_SIZE)
#define TOTAL_HEADER_SIZE			(IMAGE_HEADER_TABLE_SIZE + \
									 TOTAL_IMAGE_HEADER_SIZE + \
									 TOTAL_PARTITION_HEADER_SIZE + 64)
#define IMAGE_IDENT_OFFSET              (0x024) /**< 0x584C4E58 "XLNX" */
#define IMAGE_IDENT                     (0x584C4E58) /**< XLNX pattern */
#define IMAGE_VALID                     (0xFFFFFFFE)
#define IMAGE_HDR_COUNT_OFFSET          0x04
#define FIRST_IMAGE_HDR_OFFSET          0x0C
#define PARTITION_HDR_OFFSET            0x04
#define PARTITION_COUNT_OFFSET          0x0C
#define IMAGE_TIMESTAMP_OFFSET          0x4C
#define IMAGE_VALIDITY_WORD_OFFSET      0x60
#define FSBL_CHECKSUM_OFFSET            0x50
#define BOOTROM_HDR_CHECKSUM_WORD_COUNT 0xA
#define BOOTROM_HDR_START_OFFSET        0x20

/* Partition Header defines */
#define PARTITION_IMAGE_WORD_LEN_OFFSET	0x00	/* Word length of image */
#define PARTITION_DATA_WORD_LEN_OFFSET	0x04	/* Word length of data */
#define PARTITION_WORD_LEN_OFFSET		0x08	/* Word length of partition */
#define PARTITION_LOAD_ADDRESS_OFFSET	0x0C	/* Load addr in DDR	*/
#define PARTITION_EXEC_ADDRESS_OFFSET	0x10	/* Addr to start executing */
#define PARTITION_ADDR_OFFSET			0x14	/* Partition word offset */
#define PARTITION_ATTRIBUTE_OFFSET		0x18	/* Partition type */
#define PARTITION_HDR_CHECKSUM_OFFSET	0x3C	/* Header Checksum offset */
#define PARTITION_HDR_CHECKSUM_WORD_COUNT 0xF	/* Checksum word count */
#define PARTITION_HDR_WORD_COUNT		0x10	/* Header word len */
#define PARTITION_HDR_TOTAL_LEN			0x40	/* One partition hdr length*/

/* Attribute word defines */
#define ATTRIBUTE_IMAGE_TYPE_MASK		0xF0	/* Destination Device type */
#define ATTRIBUTE_PS_IMAGE_MASK			0x10	/* Code partition */
#define ATTRIBUTE_PL_IMAGE_MASK			0x20	/* Bit stream partition */
#define ATTRIBUTE_CHECKSUM_TYPE_MASK	0x7000	/* Checksum Type */
#define ATTRIBUTE_RSA_PRESENT_MASK		0x8000	/* RSA Signature Present */
#define ATTRIBUTE_PARTITION_OWNER_MASK	0x30000	/* Partition Owner */

#define ATTRIBUTE_PARTITION_OWNER_FSBL	0x00000	/* FSBL Partition Owner */


/**************************** Type Definitions *******************************/

typedef struct StructPartHeader {
	u32 ImageWordLen;	/* 0x0 */
	u32 DataWordLen;	/* 0x4 */
	u32 PartitionWordLen;	/* 0x8 */
	u32 LoadAddr;		/* 0xC */
	u32 ExecAddr;		/* 0x10 */
	u32 PartitionStart;	/* 0x14 */
	u32 PartitionAttr;	/* 0x18 */
	u32 SectionCount;	/* 0x1C */
	u32 CheckSumOffset;	/* 0x20 */
	u32 Pads1[1];
	u32 ACOffset;	/* 0x28 */
	u32 Pads2[4];
	u32 CheckSum;		/* 0x3C */
} PartHeader;

typedef struct StructImageHeaderTable {
    u32 Version;                    /* 0x00 */
    u32 ImageCount;                 /* x004 */
    u32 FirstPartitionHeaderOffset; /* 0x08 */
    u32 FirstImageHeaderOffset;     /* 0x0C */
    u32 AuthCertificateOffset;      /* 0x10 */
    u32 Pads[11];                   /* 0x14 */
} ImageHeaderTable;

/* PartHeaderArray */
struct HeaderArray {
	u32 Fields[16];
};

/* BootRomHeaderArray */
struct BootRomHeaderArray {
    u32 Fields[11];
};

typedef struct StructBootRomHeader {
    u32 WidthDetection;
    u32 ImageIdentification;
    u32 EncryptionStatus;
    u32 BootImageTimestamp;
    u32 SourceOffset;
    u32 LengthOfImage;
    u32 FsblLoadAddr;
    u32 StartOfExecution;
    u32 TotalImageLength;
    u32 QspiConfigWord;
    u32 HeaderChecksum;
} BootRomHeader_t;

/***************** Macros (Inline Functions) Definitions *********************/

/************************** Function Prototypes ******************************/

u32  GetPartitionCount(u8 *BootImageAddr);
void GetPartitionHeaderInfo(u8 *BootImageddr, u32 PartitionCount, PartHeader *PartitionHeader);
u32 GetBootImageTimestamp(const u8 *BootImageAddr);

time_t FlashGetBootImageTimestamp(const u32 FlashAddr);
u32 FlashIsBootImageValid(const u32 FlashAddr);
int FlashValidateBootImage (const u32 FlashAddr);
void FlashGetBootRomHeader(const u32 FlashAddr, BootRomHeader_t *BootRomHeader);
u32  FlashGetPartitionCount(const u32 FlashAddr);
void FlashGetPartitionHeaderInfo(const u32 FlashAddr, u32 PartitionCount, PartHeader *PartitionHeader);

u32  ValidateBootImage(u8 *BootImageAddr);
u32  ValidatePartitionHeader(PartHeader *PartitionHeader);
u32  ValidateFsblPartition(u8 *BootImageAddr, PartHeader *PartitionHeader);
u32  ValidatePartition(u8* BootImageAddr, PartHeader *PartitionHeader);

/************************** Variable Definitions *****************************/

#endif /* ___IMAGE_MOVER_H___ */




